@@ -37,6 +37,7 @@ enum { DEVICE_ADDED_SIGNAL, DEVICE_REMOVED_SIGNAL, LAST_SIGNAL };
 typedef struct {
 	GMainContext *main_ctx;
 	GPtrArray *devices;
+	GPtrArray *devices_removed;
 	GHashTable *dict_usb_ids;
 	GHashTable *dict_replug;
 	GThread *thread_event;
@@ -134,6 +135,7 @@ g_usb_context_dispose(GObject *object)
 
 	g_clear_pointer(&priv->main_ctx, g_main_context_unref);
 	g_clear_pointer(&priv->devices, g_ptr_array_unref);
+	g_clear_pointer(&priv->devices_removed, g_ptr_array_unref);
 	g_clear_pointer(&priv->dict_usb_ids, g_hash_table_unref);
 	g_clear_pointer(&priv->dict_replug, g_hash_table_unref);
 	g_clear_pointer(&priv->ctx, libusb_exit);
@@ -340,6 +342,11 @@ g_usb_context_remove_device(GUsbContext *self, struct libusb_device *dev)
 		return;
 	}
 
+	/* save this to a lookaside */
+	if (priv->flags & G_USB_CONTEXT_FLAGS_SAVE_EVENTS) {
+		g_ptr_array_add(priv->devices_removed, g_object_ref(device));
+	}
+
 	/* remove from enumerated list */
 	g_ptr_array_remove(priv->devices, device);
 
@@ -357,7 +364,7 @@ g_usb_context_remove_device(GUsbContext *self, struct libusb_device *dev)
 
 /**
  * g_usb_context_load:
- * @context: a #GUsbContext
+ * @self: a #GUsbContext
  * @json_object: a #JsonObject
  * @error: a #GError, or %NULL
  *
@@ -369,13 +376,39 @@ g_usb_context_remove_device(GUsbContext *self, struct libusb_device *dev)
  **/
 gboolean
 g_usb_context_load(GUsbContext *self, JsonObject *json_object, GError **error)
+{
+	return g_usb_context_load_with_tag(self, json_object, NULL, error);
+}
+
+/**
+ * g_usb_context_load_with_tag:
+ * @self: a #GUsbContext
+ * @json_object: a #JsonObject
+ * @tag: a string tag, e.g. `runtime-reload`, or %NULL
+ * @error: a #GError, or %NULL
+ *
+ * Loads any devices with a specified tag into the context from a JSON object.
+ *
+ * Return value: %TRUE on success
+ *
+ * Since: 0.4.1
+ **/
+gboolean
+g_usb_context_load_with_tag(GUsbContext *self,
+			    JsonObject *json_object,
+			    const gchar *tag,
+			    GError **error)
 {
 	GUsbContextPrivate *priv = GET_PRIVATE(self);
 	JsonArray *json_array;
 
 	g_return_val_if_fail(G_USB_IS_CONTEXT(self), FALSE);
 	g_return_val_if_fail(json_object != NULL, FALSE);
-	g_return_val_if_fail(error == NULL || *error == NULL, -1);
+	g_return_val_if_fail(error == NULL || *error == NULL, FALSE);
+
+	/* remove existing devices */
+	g_ptr_array_set_size(priv->devices, 0);
+	g_ptr_array_set_size(priv->devices_removed, 0);
 
 	if (!json_object_has_member(json_object, "UsbDevices")) {
 		g_set_error_literal(error,
@@ -392,6 +425,8 @@ g_usb_context_load(GUsbContext *self, JsonObject *json_object, GError **error)
 		    g_object_new(G_USB_TYPE_DEVICE, "context", self, NULL);
 		if (!_g_usb_device_load(device, obj_tmp, error))
 			return FALSE;
+		if (tag != NULL && !_g_usb_device_has_tag(device, tag))
+			continue;
 		g_ptr_array_add(priv->devices, g_object_ref(device));
 		g_signal_emit(self, signals[DEVICE_ADDED_SIGNAL], 0, device);
 	}
@@ -403,7 +438,7 @@ g_usb_context_load(GUsbContext *self, JsonObject *json_object, GError **error)
 
 /**
  * g_usb_context_save:
- * @context: a #GUsbContext
+ * @self: a #GUsbContext
  * @json_builder: a #JsonBuilder
  * @error: a #GError, or %NULL
  *
@@ -415,6 +450,28 @@ g_usb_context_load(GUsbContext *self, JsonObject *json_object, GError **error)
  **/
 gboolean
 g_usb_context_save(GUsbContext *self, JsonBuilder *json_builder, GError **error)
+{
+	return g_usb_context_save_with_tag(self, json_builder, NULL, error);
+}
+
+/**
+ * g_usb_context_save_with_tag:
+ * @self: a #GUsbContext
+ * @json_builder: a #JsonBuilder
+ * @tag: a string tag, e.g. `runtime-reload`, or %NULL
+ * @error: a #GError, or %NULL
+ *
+ * Saves any devices with a specified tag into an existing JSON builder.
+ *
+ * Return value: %TRUE on success
+ *
+ * Since: 0.4.1
+ **/
+gboolean
+g_usb_context_save_with_tag(GUsbContext *self,
+			    JsonBuilder *json_builder,
+			    const gchar *tag,
+			    GError **error)
 {
 	GUsbContextPrivate *priv = GET_PRIVATE(self);
 
@@ -429,8 +486,15 @@ g_usb_context_save(GUsbContext *self, JsonBuilder *json_builder, GError **error)
 	/* array of devices */
 	json_builder_set_member_name(json_builder, "UsbDevices");
 	json_builder_begin_array(json_builder);
+	for (guint i = 0; i < priv->devices_removed->len; i++) {
+		GUsbDevice *device = g_ptr_array_index(priv->devices_removed, i);
+		if (!_g_usb_device_save(device, json_builder, error))
+			return FALSE;
+	}
 	for (guint i = 0; i < priv->devices->len; i++) {
 		GUsbDevice *device = g_ptr_array_index(priv->devices, i);
+		if (tag != NULL && !_g_usb_device_has_tag(device, tag))
+			continue;
 		if (!_g_usb_device_save(device, json_builder, error))
 			return FALSE;
 	}
@@ -755,6 +819,7 @@ g_usb_context_init(GUsbContext *self)
 	priv->flags = G_USB_CONTEXT_FLAGS_NONE;
 	priv->hotplug_poll_interval = G_USB_CONTEXT_HOTPLUG_POLL_INTERVAL_DEFAULT;
 	priv->devices = g_ptr_array_new_with_free_func((GDestroyNotify)g_object_unref);
+	priv->devices_removed = g_ptr_array_new_with_free_func((GDestroyNotify)g_object_unref);
 	priv->dict_usb_ids = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
 	priv->dict_replug = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);
 
